#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <GL/glew.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include "app.h"
#include "machine.h"

bool init();
void cleanup();
void display();
void idle();
void draw_gears();
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void mouse(int bn, int st, int x, int y);
void motion(int x, int y);

static float cam_dist = 0.5;
static float cam_theta, cam_phi;
static int prev_mx, prev_my;
static bool bnstate[8];

static unsigned int start_time, prev_msec;
static Machine *machine;

int main(int argc, char **argv)
{
	glutInitWindowSize(1024, 768);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE | GLUT_MULTISAMPLE);
	glutCreateWindow("Antikythera");

	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	if(!init()) {
		return 1;
	}
	atexit(cleanup);

	glutMainLoop();
	return 0;
}

bool init()
{
	glewInit();

	glEnable(GL_MULTISAMPLE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_NORMALIZE);

	Mesh::use_custom_sdr_attr = false;

	machine = new Machine;

	const float pitch = 10.0f;

	Gear *gear1 = new Gear;
	gear1->pos = Vec3(-50, 0, 0);
	gear1->set_teeth(16, pitch);
	gear1->gen_mesh();
	machine->add_gear(gear1);

	Gear *gear2 = new Gear;
	gear2->set_teeth(32, pitch);
	gear2->pos = gear1->pos + Vec3(gear1->radius + gear2->radius - gear1->teeth_length * 0.75, 0, 0);
	gear2->thickness = 5;
	gear2->gen_mesh();
	machine->add_gear(gear2);

	Gear *gear3 = new Gear;
	gear3->set_teeth(8, pitch);
	gear3->pos = gear2->pos + Vec3(0, gear2->radius + gear3->radius - gear2->teeth_length * 0.75, 0);
	gear3->gen_mesh();
	machine->add_gear(gear3);

	machine->add_motor(0, 1.0);

	start_time = glutGet(GLUT_ELAPSED_TIME);
	return true;
}

void cleanup()
{
	delete machine;
}

void display()
{
	unsigned int msec = glutGet(GLUT_ELAPSED_TIME) - start_time;
	float dt = (float)(msec - prev_msec) / 1000.0f;
	prev_msec = msec;

	machine->update(dt);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);

	draw_gears();

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void idle()
{
	glutPostRedisplay();
}

void draw_gears()
{
	/* world scale is in meters, gears are in millimeters, scale by 1/1000 */
	glPushMatrix();
	glScalef(0.001, 0.001, 0.001);

	machine->draw();

	glPopMatrix();
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.01, 100.0);
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case 'w':
		opt_gear_wireframe = !opt_gear_wireframe;
		glutPostRedisplay();
		break;
	}
}

void mouse(int bn, int st, int x, int y)
{
	prev_mx = x;
	prev_my = y;
	bnstate[bn - GLUT_LEFT_BUTTON] = st == GLUT_DOWN;
}

void motion(int x, int y)
{
	int dx = x - prev_mx;
	int dy = y - prev_my;
	prev_mx = x;
	prev_my = y;

	if(!dx && !dy) return;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;
		cam_phi += dy * 0.5;

		if(cam_phi < -90) cam_phi = -90;
		if(cam_phi > 90) cam_phi = 90;
		glutPostRedisplay();
	}
	if(bnstate[2]) {
		cam_dist += dy * 0.01;
		if(cam_dist < 0.0) cam_dist = 0.0;
		glutPostRedisplay();
	}
}
